#include "input_handlers.h"

extern bool QUIT_VITUNES;

/*****************************************************************************
 * some misc. functions
 ****************************************************************************/

#define swap(type, x, y) \
   { type temp = x; x = y; y = temp;}

void
redraw_active_window(void)
{
   if (ui.active == ui.library)
      paint_library();
   else
      paint_playlist();
}

void
setup_viewing_playlist(playlist *p)
{
   viewing_playlist = p;

   ui.playlist->nrows   = p->nfiles;
   ui.playlist->crow    = 0;
   ui.playlist->voffset = 0;
   ui.playlist->hoffset = 0;
}


/*****************************************************************************
 * functions for working with number read from user for keybindings
 ****************************************************************************/
int _global_input_num;

void gnum_clear()    { _global_input_num = 0; }
int  gnum_get()      { return _global_input_num; }
void gnum_add(int x) { _global_input_num = _global_input_num * 10 + x; }
void gnum_set(int x) { _global_input_num = x; }


/*****************************************************************************
 * functions for setting/getting the active search direction
 ****************************************************************************/
int   _global_search_dir;
void search_dir_init()       { _global_search_dir = Forwards; }
int  search_dir_get()        { return _global_search_dir; }
void search_dir_set(int dir) { _global_search_dir = dir; }


/*****************************************************************************
 * functions for working with yank/copy buffer
 ****************************************************************************/

 /* the global yank/copy bufer */
yank_buffer _yank_buffer;

void
ybuffer_init()
{
   _yank_buffer.files = calloc(YANK_BUFFER_CHUNK_SIZE, sizeof(meta_info*));
   if (_yank_buffer.files == NULL)
      err(1, "ybuffer_init: failed to allocate");

   _yank_buffer.capacity = YANK_BUFFER_CHUNK_SIZE;
   _yank_buffer.nfiles = 0;
}

void
ybuffer_clear()
{
   _yank_buffer.nfiles = 0;
}

void
ybuffer_free()
{
   free(_yank_buffer.files);
   _yank_buffer.capacity = 0;
   _yank_buffer.nfiles = 0;
}

void
ybuffer_add(meta_info *f)
{
   meta_info **new_buff;
   int   new_capacity;

   /* do we need to realloc()? */
   if (_yank_buffer.nfiles == _yank_buffer.capacity) {
      _yank_buffer.capacity += YANK_BUFFER_CHUNK_SIZE;
      new_capacity = _yank_buffer.capacity * sizeof(meta_info*);
      if ((new_buff = realloc(_yank_buffer.files, new_capacity)) == NULL)
         err(1, "ybuffer_add: failed to realloc");

      _yank_buffer.files = new_buff;
   }

   /* add the file */
   _yank_buffer.files[ _yank_buffer.nfiles++ ] = f;
}


/*****************************************************************************
 * keybinding handlers
 ****************************************************************************/

void
load_or_play(Args a)
{
   Args dummy;
   int  idx;

   if (ui.active == ui.library) {
      /* load playlist */
      idx = ui.library->voffset + ui.library->crow;
      setup_viewing_playlist(mdb.playlists[idx]);
      paint_playlist();

      /* switch focus to that playlist */
      switch_focus(dummy);

   } else {
      /* play song */
      player_set_queue(viewing_playlist);
      playing_playlist = viewing_playlist;
      player.current = ui.active->voffset + ui.active->crow;
      player_play();
   }
}

void
pause_playback(Args a)
{
   player_pause();
}

void
seek_playback(Args a)
{
   int s;

   /* determine number of seconds to seek */
   s = 10;
   switch (a.scale) {
      case Seconds:
         s = a.amount;
         break;
      case Minutes:
         s = a.amount * 60;
         break;
      default:
         errx(1, "seek_playback: bad scale");
   }

   /* adjust for direction */
   switch (a.direction) {
      case Forwards:
         /* no change */
         break;
      case Backwards:
         s *= -1;
         break;
      default:
         errx(1, "seek_playback: bad direction");
   }

   /* seek */
   player_seek(s);
}

void
switch_focus(Args a)
{
   if (ui.active == ui.library) {
      ui.active = ui.playlist;
      if (ui.lhide) {
         ui_hide_library();
         paint_all();
      }
   } else {
      ui.active = ui.library;
      if (ui.lhide) {
         ui_unhide_library();
         paint_all();
      }
   }

   paint_status_bar();
}

void
redraw(Args a)
{
   ui_clear();
   paint_all();
}

void
enter_cmd_mode(Args a)
{
   bool   found;
   char  *cmd;
   char **argv;
   int    argc;
   int    i;

   /* get command from user */
   if (ui_command_get_string(":", &cmd) == UI_USER_CANCELLED) {
      werase(ui.command);
      wrefresh(ui.command);
      return;
   }

   /* convert to argc/argv structure */
   switch(str2argv(cmd, &argc, &argv)) {
      case PARGV_OK:
         break;
      case PARGV_ERR_ESCAPE:
         paint_error("parse error: unused escape sequence in '%s'", cmd);
         return;
      case PARGV_ERR_QUOTES:
         paint_error("parse error: mis-matched quotes '%s'", cmd);
         return;
      default:
         errx(1, "enter_cmd_mode: unknown return code from str2argv");
   }

   /* search path for appropriate command to execute */
   found = false;
   for (i = 0; i < CommandPathSize && !found; i++) {
      if (strcmp(argv[0], CommandPath[i].name) == 0) {
         found = true;
         (CommandPath[i].func)(argc, argv);
      }
   }

   if (!found)
      paint_error("Unknown command '%s'.", argv[0]);

   argv_free(&argc, &argv);
}

void
scroll_row(Args a)
{
   int n;

   /* determine how many rows to scroll */
   n = 1;
   if (gnum_get() > 0) {
      n = gnum_get();
      gnum_clear();
   }

   /* update current row */
   switch (a.direction) {
      case Down:
         ui.active->crow += n;
         break;
      case Up:
         ui.active->crow -= n;
         break;
      default:
         errx(1, "bad dir in scroll_row");
   }

   /* handle off-the-edge cases */
   if (ui.active->crow < 0) {
      swindow_scroll(ui.active, Up, -1 * ui.active->crow);
      ui.active->crow = 0;
   }

   if (ui.active->voffset + ui.active->crow >= ui.active->nrows)
      ui.active->crow = ui.active->nrows - ui.active->voffset - 1;

   if (ui.active->crow >= ui.active->h) {
      swindow_scroll(ui.active, Down, ui.active->crow - ui.active->h + 1);
      ui.active->crow = ui.active->h - 1;
   }

   if (ui.active->crow < 0)
      ui.active->crow = 0;

   /* redraw */
   redraw_active_window();
   paint_status_bar();
}

void
scroll_col(Args a)
{
   int maxhoff;
   int maxlen;
   int n;
   int i;

   /* determine how many cols to scroll */
   n = 1;
   if (gnum_get() > 0) {
      n = gnum_get();
      gnum_clear();
   }

   /* determine maximum horizontal offset */
   maxhoff = 0;
   if (ui.active == ui.playlist) {
      if (mi_display_getwidth() > ui.active->w)
         maxhoff = mi_display_getwidth() - ui.active->w;
   } else {
      maxlen = 0;
      for (i = 0; i < mdb.nplaylists; i++) {
         if (strlen(mdb.playlists[i]->name) > maxlen)
            maxlen = strlen(mdb.playlists[i]->name);
      }
      if (maxlen > ui.active->w)
         maxhoff = maxlen - ui.active->w;
   }

   /* scroll */
   switch (a.amount) {
      case One:
         swindow_scroll(ui.active, a.direction, n);
         if (ui.active->hoffset > maxhoff)
            ui.active->hoffset = maxhoff;
         break;
      case Whole:
         switch (a.direction) {
            case Left:
               ui.active->hoffset = 0;
               break;
            case Right:
               ui.active->hoffset = maxhoff;
               break;
            default:
               errx(1, "scroll_col: bad direction");
         }
         break;
      default:
         errx(1, "scroll_col: bad amount");
   }

   /* redraw */
   redraw_active_window();
   paint_status_bar();
}

void
scroll_page(Args a)
{
   bool maintain_row_idx;
   int  voffset_original;
   int  max_row;
   int  diff;
   int  n;

   voffset_original = ui.active->voffset;

   /* determine max row number */
   if (ui.active->voffset + ui.active->h >= ui.active->nrows)
      max_row = ui.active->nrows - ui.active->voffset - 1;
   else
      max_row = ui.active->h - 1;

   /* determine how many pages to scroll */
   n = 1;
   if (gnum_get() > 0) {
      n = gnum_get();
      gnum_clear();
   }

   /* determine how much crow should change */
   diff = 1;
   maintain_row_idx = false;
   switch (a.amount) {
      case One:
         diff = 1 * n;
         maintain_row_idx = true;
         break;
      case Half:
         diff = (ui.active->h / 2) * n;
         maintain_row_idx = false;
         break;
      case Whole:
         diff = ui.active->h * n;
         maintain_row_idx = false;
         break;
      default:
         errx(1, "bad amount in scroll_page");
   }
   swindow_scroll(ui.active, a.direction, diff);

   /* handle updating the current row */
   if (maintain_row_idx) {
      if (a.direction == Down) {
         ui.active->crow -= diff;
         if (ui.active->crow < 0)
            ui.active->crow = 0;
      } else {
         ui.active->crow += diff;
         if (ui.active->crow >= ui.active->h)
            ui.active->crow = ui.active->h - 1;
      }

   } else {
      /* we only move current row here if there was a change this time
       * in the vertical offset */

      if (a.direction == Down
      &&  ui.active->voffset == voffset_original) {
         ui.active->crow += diff;
         if (ui.active->crow > max_row)
            ui.active->crow = max_row;
      }

      if (a.direction == Up
      &&  ui.active->voffset == 0 && voffset_original == 0) {
         ui.active->crow -= diff;
         if (ui.active->crow < 0)
            ui.active->crow = 0;
      }
   }

   redraw_active_window();
   paint_status_bar();
}

void
jumpto_page(Args a)
{
   int n;
   int max_row;

   /* get offset */
   n = 0;
   if (gnum_get() > 0) {
      n = gnum_get();
      gnum_clear();
   }

   /* determine max row number */
   if (ui.active->voffset + ui.active->h >= ui.active->nrows)
      max_row = ui.active->nrows - ui.active->voffset - 1;
   else
      max_row = ui.active->h - 1;

   /* update current row */
   switch (a.location) {
      case Top:
         ui.active->crow = n - 1;
         break;
      case Middle:
         ui.active->crow = max_row / 2;
         break;
      case Bottom:
         ui.active->crow = max_row - n + 1;
         break;
      default:
         errx(1, "bad loc in jumpto_page");
   }

   /* sanitize current row */
   if (ui.active->crow < 0)
      ui.active->crow = 0;

   if (ui.active->crow >= ui.active->h)
      ui.active->crow = ui.active->h - 1;

   if (ui.active->crow > max_row)
      ui.active->crow = max_row;

   /* redraw */
   redraw_active_window();
   paint_status_bar();
}

void
jumpto_file(Args a)
{
   float pct;
   int   n, line;

   /* determine line/percent to jump to */
   n = -1;
   if (gnum_get() > 0) {
      n = gnum_get();
      gnum_clear();
   }

   /* get line number to jump to */
   line = -1;
   switch (a.amount) {
      case Number:
         if (n < 0 || n >= ui.active->nrows)
            line = ui.active->nrows;
         else
            line = n;

         break;

      case Percent:
         if (n <= 0 || n > 100)
            return;

         pct = (float) n / 100;
         line = pct * (float) ui.active->nrows;
         break;

      default:
         errx(1, "jumpto_file: np parameter is bad");
   }

   /* jump */
   if (line <= ui.active->voffset) {
      swindow_scroll(ui.active, Up, ui.active->voffset - line + 1);
      ui.active->crow = 0;
   } else if (line > ui.active->voffset + ui.active->h) {
      swindow_scroll(ui.active, Down, line - (ui.active->voffset + ui.active->h));
      ui.active->crow = ui.active->h - 1;
   } else {
      ui.active->crow = line - ui.active->voffset - 1;
   }

   /* redraw */
   redraw_active_window();

   if (a.location != -1)   /* XXX a damn dirty hack for now. see search_find */
      paint_status_bar();
}

void
search(Args a)
{
   Args   find_args;
   char  *search_phrase;
   char  *prompt;
   char **argv  = NULL;
   int    argc = NULL;
   int    i;

   /* TODO yeah, support searching in the library window */
   if (ui.active == ui.library) {
      paint_error("Sorry, you can't search in library window yet :(");
      return;
   }

   /* determine prompt to use */
   prompt = "uh-oh";
   switch (a.direction) {
      case Forwards:
         prompt = "/";
         break;
      case Backwards:
         prompt = "?";
         break;
      default:
         errx(1, "bad dir to search");
   }

   /* get search phrase from user */
   if (ui_command_get_string(prompt, &search_phrase) == UI_USER_CANCELLED) {
      wclear(ui.command);
      wrefresh(ui.command);
      return;
   }

   /* set the global query description and the search direction */
   switch (str2argv(search_phrase, &argc, &argv)) {
      case PARGV_OK:
         mi_query_clear();
         for (i = 0; i < argc; i++)
            mi_query_add_token(argv[i]);

         mi_query_setraw(search_phrase);
         search_dir_set(a.direction);
         argv_free(&argc, &argv);
         break;

      case PARGV_ERR_ESCAPE:
         paint_error("parse error: unused escape in '%s", search_phrase);
         return;

      case PARGV_ERR_QUOTES:
         paint_error("parse error: mis-matched quotes in '%s", search_phrase);
         return;

      default:
         errx(1, "search: str2argv returned unknown error");
   }

   /* do the search */
   find_args.direction = Same;
   search_find(find_args);
}

void
search_find(Args a)
{
   Args  foo;
   char *msg;
   int   dir;
   int   start_idx;
   int   idx;
   int   c;

   /* TODO yeah, support searching in the library window */
   if (ui.active == ui.library) {
      paint_error("Sorry, you can't search in library window yet :(");
      return;
   }

   /* determine direction to do the search */
   dir = Forwards;
   switch (a.direction) {
      case Same:
         dir = search_dir_get();
         break;

      case Opposite:
         if (search_dir_get() == Forwards)
            dir = Backwards;
         else
            dir = Forwards;
         break;

      default:
         errx(1, "search_find: bad direction");
   }

   /* start looking from current row */
   start_idx = ui.active->voffset + ui.active->crow;
   msg = NULL;
   for (c = 1; c < viewing_playlist->nfiles + 1; c++) {

      /* get idx of record */
      if (dir == Forwards)
         idx = start_idx + c;
      else
         idx = start_idx - c;

      /* normalize idx */
      if (idx < 0) {
         idx = viewing_playlist->nfiles + idx;
         msg = "search hit BOTTOM, continuing at TOP";

      } else if (idx >= viewing_playlist->nfiles) {
         idx %= viewing_playlist->nfiles;
         msg = "search hit TOP, continuing at BOTTOM";
      }

      /* check if record at idx matches */
      if (mi_match(viewing_playlist->files[idx])) {
         /* found one, jump to it */
         if (msg != NULL)
            paint_message(msg);

         gnum_set(idx + 1);
         foo.amount = Number;
         foo.location = -1;
         jumpto_file(foo);
         return;
      }
   }

   paint_error("Pattern not found: %s", mi_query_getraw());
}


/*****************************************************************************
 * yank, cut, and paste keybindings.
 ****************************************************************************/

void
yank(Args a)
{
   bool got_target;
   int  start, end;
   int  input;
   int  n;

   if (ui.active == ui.library) {
      paint_error("cannot yank in library window");
      return;
   }

   if (viewing_playlist->nfiles == 0) {
      paint_error("nothing to yank!");
      return;
   }

   /* determine range */
   n = 1;
   if (gnum_get() > 0) {
      n = gnum_get();
      gnum_clear();
   }

   /* get next input from user */
   got_target = false;
   start = 0;
   end = 0;
   while ((input = getch()) && !got_target) {
      if (input == ERR)
         continue;

      switch (input) {
         case 'y':   /* yank next n lines */
            start = ui.active->voffset + ui.active->crow;
            end = start + n;
            got_target = true;
            break;

         case 'G':   /* yank to end of current playlist */
            start = ui.active->voffset + ui.active->crow;
            end = ui.active->nrows;
            got_target = true;
            break;

         /* TODO handle other directions ( j/k/H/L/M/^u/^d/ etc. )
          * here.  this will be ... tricky.
          * might want to re-organize other stuff?
          */

         default:
            ungetch(input);
            return;
      }
   }

   /* sanitize start and end */
   if (end > ui.active->nrows)
      end = ui.active->nrows;

   /* clear existing yank buffer and add new stuff */
   ybuffer_clear();
   for (n = start; n < end; n++)
      ybuffer_add(viewing_playlist->files[n]);

   /* notify user # of rows yanked */
   paint_message("Yanked %d files.", end - start);
}

/* TODO so much of the logic in cut() is similar with that of yank() above.
 * combine the two, use an Args parameter to determine if the operation is
 * a yank or a delete
 */
void
cut(Args a)
{
   playlist *p;
   char *warning;
   bool  got_target;
   int   start, end;
   int   response;
   int   input;
   int   n;

   /* determine range */
   n = 1;
   if (gnum_get() > 0) {
      n = gnum_get();
      gnum_clear();
   }

   /* get range */
   got_target = false;
   start = 0;
   end = 0;
   while ((input = getch()) && !got_target) {
      if (input == ERR)
         continue;

      switch (input) {
         case 'd':   /* delete next n lines */
            start = ui.active->voffset + ui.active->crow;
            end = start + n;
            got_target = true;
            break;

         case 'G':   /* delete to end of current playlist */
            start = ui.active->voffset + ui.active->crow;
            end = ui.active->nrows;
            got_target = true;
            break;

         default:
            ungetch(input);
            return;
      }
   }

   /* handle deleting of a whole playlist */
   if (ui.active == ui.library) {
      /* get playlist to delete */
      n = ui.active->voffset + ui.active->crow;
      p = mdb.playlists[n];

      /* i simply do NOT want to be able to delete multiple playlists while
       * drunk. */
      if (end != start + 1) {
         paint_error("cannot delete multiple playlists");
         return;
      }
      if (p == mdb.library || p == mdb.filter_results) {
         paint_error("cannot delete pseudo-playlists like LIBRARY or FILTER");
         return;
      }

      asprintf(&warning, "Are you sure you want to delete '%s'?", p->name);
      if (warning == NULL)
         err(1, "cut: failed to asprintf");

      /* make sure user wants this */
      if (ui_command_get_yesno(warning, &response) == UI_USER_CANCELLED) {
         paint_message("delete of '%s' cancelled", p->name);
         return;
      }
      if (response != 1) {
         paint_message("playlist '%s' not deleted", p->name);
         return;
      }

      /* delete playlist and redraw library window */
      if (viewing_playlist == p) {
         setup_viewing_playlist(mdb.library);
         paint_playlist();
      }
      ui.library->nrows--;
      medialib_playlist_remove(n);
      paint_library();
      free(warning);
      return;
   }

   /* return to handling of deleting files in a playlist... */

   /* sanitize start and end */
   if (end > ui.active->nrows)
      end = ui.active->nrows;

   /* clear existing yank buffer and add new stuff */
   ybuffer_clear();
   for (n = start; n < end; n++)
      ybuffer_add(viewing_playlist->files[n]);

   /* delete files */
   for (n = start; n < end; n++)
      playlist_file_remove(viewing_playlist, start);


   /* update ui appropriately */
   viewing_playlist->needs_saving = true;
   ui.active->nrows = viewing_playlist->nfiles;
   if (ui.active->voffset + ui.active->crow >= ui.active->nrows)
      ui.active->crow = ui.active->nrows - ui.active->voffset - 1;


   /* redraw */
   paint_playlist();
   paint_message("%d fewer files.", end - start);
}

void
paste(Args a)
{
   int start;
   int i;

   if (ui.active == ui.library) {
      paint_error("currently cannot paste in library window.  :(");
      return;
      /* XXX it might be reasonable to just past to the end of the currently
       * selected playlist here... this would be easy.  chat with others
       * first...see what they think.
       */
   }

   if (_yank_buffer.nfiles == 0) {
      paint_error("nothing to paste");
      return;
   }

   /* determine index in playlist to insert new files after */
   start = 0;
   switch (a.location) {
      case Before:
         start = ui.active->voffset + ui.active->crow;
         break;
      case After:
         start = ui.active->voffset + ui.active->crow + 1;
         if (start > viewing_playlist->nfiles)
            start = viewing_playlist->nfiles;

         break;
      default:
         errx(1, "paste: bad location");
   }

   /* add files from buffer */
   for (i = 0; i < _yank_buffer.nfiles; i++)
      playlist_file_add(viewing_playlist, _yank_buffer.files[i], start + i);

   /* update ui */
   ui.active->nrows += i;
   viewing_playlist->needs_saving = true;

   /* redraw */
   paint_playlist();
   paint_message("Pasted %d files.", i);
}


/*****************************************************************************
 * command mode handlers
 * XXX NOTE XXX:
 * All of these functions paint_error/message something appropriate when
 * things go wrong AND THEY SHOULD ALSO return a negative value.  This is
 * because the "load_config" routine in vitunes.c depends on this to know if
 * any of the commands in a configuration file fail.
 ****************************************************************************/

int
cmd_quit(int argc, char *argv[])
{
   bool forced;
   int  i;

   if (argc != 1) {
      paint_error("usage: q[!]");
      return -1;
   }

   /* is this a forced quit? */
   forced = (strcmp(argv[0], "q!") == 0);

   /* check if there are any unsaved changes if not forced */
   if (!forced) {
      for (i = 0; i < mdb.nplaylists; i++) {
         if (mdb.playlists[i]->needs_saving) {
            paint_error("there are playlists with unsaved changes.  use \"q!\" to force.");
            return -1;
         }
      }
   }

   QUIT_VITUNES = true;
   return 0;
}

int
cmd_write(int argc, char *argv[])
{
   playlist *dup;
   char     *filename;
   bool      forced;
   bool      will_clobber;
   int       i, clobber_index;

   if (argc > 2) {
      paint_error("usage: w[!] [name]");
      return -1;
   }

   forced = (strcmp(argv[0], "w!") == 0);

   if (argc == 1) {  /* "save" */

      /* can't save library or filter results */
      if (viewing_playlist == mdb.library
      ||  viewing_playlist == mdb.filter_results) {
         paint_error("use \"w name\" when saving psuedo-playlists like library/filter");
         return -1;
      }

      /* can't save a new playlist that has no name */
      if (viewing_playlist->filename == NULL) {
         paint_error("use \"w name\" for new playlists");
         return -1;
      }

      /* do the save... */
      playlist_save(viewing_playlist);
      viewing_playlist->needs_saving = false;
      paint_message("\"%s\" %d songs written",
         viewing_playlist->filename, viewing_playlist->nfiles);

   } else { /* "save as" */

      /* build filename for playlist */
      asprintf(&filename, "%s/%s.playlist", mdb.playlist_dir, argv[1]);
      if (filename == NULL)
         err(1, "cmd_write: failed to allocate space for filename");

      /* check to see if playlist with that name already exists */
      will_clobber = false;
      clobber_index = -1;
      for (i = 0; i < mdb.nplaylists; i++) {
         if (mdb.playlists[i]->filename != NULL
         &&  strcmp(mdb.playlists[i]->filename, filename) == 0) {
            will_clobber = true;
            clobber_index = i;
         }
      }

      if (will_clobber && !forced) {
         paint_error("playlist with that name exists (use \"w!\" to overwrite)");
         free(filename);
         return -1;
      }

      /* if reached here, we're going to do the save-as... */

      /* duplicate playlist */
      dup = playlist_dup(viewing_playlist, filename, argv[1]);
      if (will_clobber)
         medialib_playlist_remove(clobber_index);
      else
         ui.library->nrows++;

      /* TODO
       * If the original playlist was a new one, with no name (name == NULL)
       * then we should remove that playlist from the medialib and display
       * here.
       */

      /* do the save-as... */
      playlist_save(dup);
      medialib_playlist_add(dup);

      dup->needs_saving = false;
      viewing_playlist->needs_saving = false;

      paint_library();
      paint_message("\"%s\" %d songs written",
         filename, viewing_playlist->nfiles);
   }

   return 0;
}

int
cmd_mode(int argc, char *argv[])
{
   if (argc != 2) {
      paint_error("usage: mode [ linear | loop | random ]");
      return -1;
   }

   if (strcasecmp(argv[1], "linear") == 0)
      player.mode = MP_MODE_LINEAR;
   else if (strcasecmp(argv[1], "loop") == 0)
      player.mode = MP_MODE_LOOP;
   else if (strcasecmp(argv[1], "random") == 0)
      player.mode = MP_MODE_RANDOM;
   else {
      paint_error("invalid mode \"%s\".  must be one of: linear, loop, or random", argv[1]);
      return -1;
   }

   paint_message("mode changed to: %s", argv[1]);
   return 0;
}

int
cmd_new(int argc, char *argv[])
{
   playlist *p;
   char     *name;
   int       i;

   if (argc > 2) {
      paint_error("usage: new [name]");
      return -1;
   }

   /* determine what name to use (if one was provided) */
   if (argc == 2) {
      /* a name was provided.  check to see if such a playlist exists already */
      for (i = 0; i < mdb.nplaylists; i++) {
         if (strcmp(mdb.playlists[i]->name, argv[1]) == 0) {
            paint_error("playlist \"%s\" already exists.", argv[1]);
            return -1;
         }
      }

      name = argv[1];
   } else
      name = "untitled";

   /* create playlist and set name */
   p = playlist_new();
   p->needs_saving = true;
   if ((p->name = strdup(name))== NULL)
      err(1, "cmd_new: strdup failed");

   /* add playlist to media library and update ui */
   medialib_playlist_add(p);
   ui.library->nrows++;

   /* redraw */
   paint_library();
   paint_message("playlist \"%s\" added", name);

   return 0;
}

int
cmd_filter(int argc, char *argv[])
{
   playlist *results;
   char     *search_phrase;
   bool      match;
   int       i;

   if (argc == 1) {
      paint_error("usage: filter[!] <query>");
      return -1;
   }

   /* determine what kind of filter we're doing */
   match = argv[0][strlen(argv[0]) - 1] != '!';

   /* clear existing global search description */
   mi_query_clear();

   /* unravel the search phrase and set the raw-query */
   search_phrase = argv2str(argc - 1, argv + 1);
   mi_query_setraw(search_phrase);
   free(search_phrase);

   /* clear existing and setup global query description */
   for (i = 1; i < argc; i++)
      mi_query_add_token(argv[i]);

   /* filter current playlist */
   results = playlist_filter(viewing_playlist, match);

   /* swap out necessary bits of the filter results playlists with this new one */
   swap(meta_info **, results->files,    mdb.filter_results->files);
   swap(int, results->nfiles,   mdb.filter_results->nfiles);
   swap(int, results->capacity, mdb.filter_results->capacity);
   playlist_free(results);

   /* redraw */
   setup_viewing_playlist(mdb.filter_results);
   paint_library();
   paint_playlist();

   return 0;
}

int
cmd_sort(int argc, char *argv[])
{
   if (argc != 2) {
      paint_error("usage: sort <sort-description>");
      return -1;
   }

   /* setup global sort description */
   if (mi_sort_set(argv[1]) != MI_OK) {
      paint_error("bad sort description");
      return -1;
   }

   /* do the actual sort */
   qsort(viewing_playlist->files, viewing_playlist->nfiles,
      sizeof(meta_info*), mi_compare);

   /* redraw */
   paint_playlist();

   return 0;
}

int
cmd_display(int argc, char *argv[])
{
   if (argc != 2) {
      paint_error("usage: display [ reset | <display-description> ]");
      return -1;
   }

   /* reset existing display? */
   if (strcasecmp(argv[1], "reset") == 0) {
      mi_display_reset();
   } else {
      /* no.  setup global display description */
      if (mi_display_set(argv[1]) != MI_OK) {
         paint_error("%s: bad display description", argv[0]);
         return -1;
      }
   }

   /* redraw */
   paint_playlist();

   return 0;
}

int
cmd_color(int argc, char *argv[])
{
   char  *item;
   char  *fg, *bg;
   int    i_item, i_fg, i_bg, j;

   if (argc != 2) {
      paint_error("usage: %s ITEM=FG,BG", argv[0]);
      return -1;
   }

   /* extract item and foreground/background colors */
   item = argv[1];
   if ((fg = strchr(item, '=')) == NULL) {
      paint_error("usage: %s ITEM=FG,BG", argv[0]);
      return -1;
   }
   *fg = '\0';
   fg++;

   if ((bg = strchr(fg, ',')) == NULL) {
      paint_error("usage: %s ITEM=FG,BG", argv[0]);
      return -1;
   }
   *bg = '\0';
   bg++;

   /* convert all */
   if ((i_item = paint_str2item(item)) < 0) {
      paint_error("invalid item '%s'", item);
      return -1;
   }

   if ((i_fg = paint_str2color(fg)) < 0) {
      paint_error("invalid foreground color '%s'", fg);
      return -1;
   }

   if ((i_bg = paint_str2color(bg)) < 0) {
      paint_error("invalid background color '%s'", bg);
      return -1;
   }

   /* init color */
   init_pair(i_item, i_fg, i_bg);

   /* if this was a cinfo item, indicate that it was set */
   for (j = 0; j < MI_NUM_CINFO; j++) {
      if (i_item == colors.cinfos[j])
         colors.cinfos_set[j] = true;
   }

   /* redraw */
   ui_clear();
   paint_all();

   return 0;
}

int
cmd_set(int argc, char *argv[])
{
   const char *err;
   char *property;
   char *value;
   int   max_w, max_h, new_width;   /* lwidth */

   if (argc != 2) {
      paint_error("usage: %s <property>=<value>", argv[0]);
      return -1;
   }

   /* extract property and value */
   property = argv[1];
   if ((value = strchr(property, '=')) == NULL) {
      paint_error("usage: %s <property>=<value>", argv[0]);
      return -1;
   }
   *value = '\0';
   value++;

   /* handle property */

   if (strcasecmp(property, "lwidth") == 0) {
      /* get max width and height */
      getmaxyx(stdscr, max_h, max_w);

      /* validate and convert width user provided */
      new_width = (int)strtonum(value, 1, max_w, &err);
      if (err != NULL) {
         paint_error("%s %s: bad width: '%s' %s",
            argv[0], property, value, err);
         return -1;
      }

      /* redraw */
      ui.lwidth = new_width;
      ui_resize();
      ui_clear();
      paint_all();

   } else if (strcasecmp(property, "lhide") == 0) {
      if (strcasecmp(value, "true") == 0) {
         ui.lhide = true;
         if (ui.active == ui.playlist) ui_hide_library();
         paint_all();
         paint_message("library window hidden");
      } else if (strcasecmp(value, "false") == 0) {
         ui.lhide = false;
         if (ui.library->cwin == NULL) ui_unhide_library();
         paint_all();
         paint_message("library window un-hidden");
      } else {
         paint_error("%s %s: value must be one of true/false",
            argv[0], property);
         return -1;
      }

   } else if (strcasecmp(property, "match-fname") == 0) {
      if (strcasecmp(value, "true") == 0) {
         mi_query_match_filename = true;
         paint_message("filenames will be matched against");
      } else if (strcasecmp(value, "false") == 0) {
         mi_query_match_filename = false;
         paint_message("filenames will NOT be matched against");
      } else {
         paint_error("%s %s: value must be one of true/false",
            argv[0], property);
         return -1;
      }

   } else {
      paint_error("%s: unknown property '%s'", argv[0], property);
      return -1;
   }

   return 0;
}

int
cmd_reload(int argc, char *argv[])
{
   char *db_file;
   char *playlist_dir;

   if (argc != 2) {
      paint_error("usage: %s [ db | conf ]", argv[0]);
      return -1;
   }

   /* reload database or config file */
   if (strcasecmp(argv[1], "db") == 0) {

      db_file = strdup(mdb.db_file);
      playlist_dir = strdup(mdb.playlist_dir);
      if (db_file == NULL || playlist_dir == NULL)
         err(1, "cmd_reload: failed to strdup db/playlist info");

      /* stop playback TODO investigate a nice way around this */
      player_stop();

      /* reload db */
      medialib_destroy();
      medialib_load(db_file, playlist_dir);

      free(db_file);
      free(playlist_dir);

      /* re-setup ui basics */
      playing_playlist = NULL;
      setup_viewing_playlist(mdb.library);
      ui.library->voffset = 0;
      ui.library->nrows = mdb.nplaylists;
      ui.library->crow = 0;
      paint_all();

   } else if (strcasecmp(argv[1], "conf") == 0)
      load_config();
   else {
      paint_error("usage: %s [ db | conf ]", argv[0]);
      return -1;
   }

   return 0;
}
