#ifndef INPUT_HANDLERS_H
#define INPUT_HANDLERS_H

#include "paint.h"
#include "str2argv.h"
#include "vitunes.h"


/*****************************************************************************
 * functions for working with number read from user for keybindings
 ****************************************************************************/
void gnum_clear();
int  gnum_get();
void gnum_add(int x);


/*****************************************************************************
 * functions for working with active search direction
 ****************************************************************************/
void search_dir_init();
void search_dir_set(int dir);
int  search_dir_get();


/*****************************************************************************
 * functions for working with yank/copy buffer
 ****************************************************************************/
#define YANK_BUFFER_CHUNK_SIZE 100
typedef struct {
   meta_info   **files;
   int           nfiles;
   int           capacity;
} yank_buffer;

void ybuffer_init();
void ybuffer_clear();
void ybuffer_free();
void ybuffer_add(meta_info *f);


/* 2 handy functions used frequently */
void setup_viewing_playlist(playlist *p);
void redraw_active_window();

/*****************************************************************************
 * keybinding handlers
 ****************************************************************************/
typedef struct {
   int   direction;
   int   location;
   int   amount;
   int   scale;
} Args;

typedef struct {
   int   keycode;
   void  (*func)(Args);
   Args  args;
} keybinding;

extern const keybinding KeyBindings[];
extern const int KeyBindingsSize;

void load_or_play(Args a);
void pause_playback(Args a);
void seek_playback(Args a);
void switch_focus(Args a);
void redraw(Args a); 
void enter_cmd_mode(Args a);
void scroll_row(Args a);
void scroll_col(Args a);
void scroll_page(Args a);
void jumpto_file(Args a);
void jumpto_page(Args a);
void yank(Args a);
void paste(Args a);
void cut(Args a);
void search(Args a);
void search_find(Args a);


/*****************************************************************************
 * command-mode handlers
 ****************************************************************************/
/* used for command-mode handlers */
typedef struct {
   char  *name;
   int   (*func)(int argc, char **argv);
} cmd;

extern const cmd CommandPath[];
extern const int CommandPathSize;

int  cmd_quit(int argc, char *argv[]);
int  cmd_write(int argc, char *argv[]);
int  cmd_mode(int argc, char *argv[]);
int  cmd_new(int argc, char *argv[]);
int  cmd_filter(int argc, char *argv[]);
int  cmd_sort(int argc, char *argv[]);
int  cmd_display(int argc, char *argv[]);
int  cmd_color(int argc, char *argv[]);
int  cmd_set(int argc, char *argv[]);
int  cmd_reload(int argc, char *argv[]);

#endif
