/* TODO Maintain an index of the filenames of each meta_info in the db
 * and provide ability to bin-search against that when loading
 * playlists from file.  As-is, really large playlists can take a
 * while.
 */

#ifndef MEDIA_LIBRARY_H
#define MEDIA_LIBRARY_H

#include <sys/errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fts.h>

#include "meta_info.h"
#include "playlist.h"
#include "debug.h"

#define MEDIALIB_PLAYLISTS_CHUNK_SIZE  100

/* current database file-format version */
#define DB_VERSION_MAJOR   1
#define DB_VERSION_MINOR   0
#define DB_VERSION_OTHER   1

typedef struct
{
   /* some locations of where things are loaded/saved */
   char     *db_file;      /* file containing the database */
   char     *playlist_dir; /* directory where playlists are stored */

   /* psuedo-playlists */
   playlist *library;         /* playlist representing the database */
   playlist *filter_results;  /* playlist representing results of a filter */
      /* NOTE: these also exist as the first two members of the playlists
       * array below (this makes searching playlists, displaying, etc.
       * easier) */

   /* the playlists */
   playlist **playlists;            /* array of all playlists */
   int        nplaylists;           /* num playlists in array */
   int        playlists_capacity;   /* total size of playlists array */


} medialib;


/* the global medialib object used throughout vitunes */
extern medialib mdb;


/* load/free the global media library */
void medialib_load(const char *db_file, const char *playlist_dir);
void medialib_destroy();

/* add/remove playlists to/from the global media library */
void medialib_playlist_add(playlist *p);
void medialib_playlist_remove(int pindex);

/* create all the necessary files/directories for vitunes medialib */
void medialib_setup_files(const char *vitunes_dir, const char *db_file,
   const char *playlist_dir);

/* load/save the core database from/to disk */
void medialib_db_load(const char *db_file);
void medialib_db_save(const char *db_file);

/* update/add files to the database */
void medialib_db_update();
void medialib_db_scan_dirs(char *dirlist[]);

/* flush the database in a semi-human readable format to output */
void medialib_db_flush(FILE *out);

#endif
