/*
 * Copyright (c) 2010 Ryan Flannery <ryan.flannery@gmail.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef PAINTING_H
#define PAINTING_H

#include <curses.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <time.h>

#include "uinterface.h"
#include "meta_info.h"
#include "playlist.h"
#include "player.h"
#include "vitunes.h"

/* colors used by paint - each of these will be a number for a COLOR_PAIR */
typedef struct {
   int   bars;
   int   tildas;

   /* error/info messages */
   int   errors;
   int   messages;

   /* windows */
   int   player;
   int   status;
   int   library;
   int   playlist;

   /* currently playing rows in both lib & playlist windows */
   int   playing;

   /* fields in the playlist window */
   int   cinfos[MI_NUM_CINFO];
   bool  cinfos_set[MI_NUM_CINFO];

} _colors;
extern _colors colors;

/* routines for painting each window */
void paint_status_bar();
void paint_player();
void paint_library();
void paint_playlist();
void paint_all();

/* routines for painting errors/messages in the command/status window */
void paint_error(char *fmt, ...);
void paint_message(char *fmt, ...);

/* for setting up and working with the colors */
void paint_setup_colors();
int  paint_str2item(const char *str);
int  paint_str2color(const char *str);

#endif
