/*
 * Copyright (c) 2010 Ryan Flannery <ryan.flannery@gmail.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef MEDIA_PLAYER_H
#define MEDIA_PLAYER_H

#include <sys/wait.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>

#include "meta_info.h"
#include "playlist.h"
#include "debug.h"

/* XXX old, old code lurks here.  you have been warned */


/* available play-modes.
 *    Linear:  Songs in the queue play in the order they appear
 *    Loop:    Like linear, but when the end is reached, the queue restarts
 *    Random:  Songs are chosen at random and play never ends
 */
typedef enum
{  MP_MODE_LINEAR,
   MP_MODE_LOOP,
   MP_MODE_RANDOM
} playmode;


/* structure used to track status of media player */
typedef struct {
   bool  is_playing;  /* flag indicating if media player is playing */
   float position;    /* position, in seconds, into currently playing file */
} mp_status_t;

/* global status object used to track the status of the media player */
extern mp_status_t mp_status;


/* main structure encapsulating info about the media player itself */
typedef struct
{
   /* info about the forked media player */
   pid_t    pid;           /* process id */
   int      pipe_read;     /* read-end of pipe */
   int      pipe_write;    /* write-end of pipe */

   /* currently playing queue information */
   playlist *queue;     /* pointer to size of queue */
   int       current;   /* index into above playlist files of currently
                         * playing file
                         */

   /* flags used to track state of the media player */
   bool     should_be_playing;
   bool     paused;
   bool     just_started_playing;

   /* misc info about the player */
   playmode mode;
   int      rand_seed;

} media_player;

/* global media player object */
extern media_player player;


/* setup/destory global media player */
void player_init();
void player_shutdown();

/* media player control functions */
void player_set_queue(playlist *queue);
void player_play();
void player_stop();
void player_pause();
void player_seek(int seconds);

/* monitor function used to update the status of the player.  should be
 * called from a timer
 */
void player_monitor();

/* some misc. functions */
int  get_next_song_index();


#endif
